/*-
 *   Copyright (C) 2011 by Maxim Ignatenko
 *   gelraen.ua@gmail.com
 *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *    modification, are permitted provided that the following conditions   *
 *    are met:                                                             *
 *     * Redistributions of source code must retain the above copyright    *
 *       notice, this list of conditions and the following disclaimer.     *
 *     * Redistributions in binary form must reproduce the above copyright *
 *       notice, this list of conditions and the following disclaimer in   *
 *       the documentation and/or other materials provided with the        *
 *       distribution.                                                     *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/endian.h>

#include <netgraph/ng_message.h>
#include <netgraph/ng_parse.h>
#include <netgraph/netgraph.h>

#include <sys/ctype.h>
#include <netinet/in.h>
#include <netinet/ip.h>

#include "ng_mikrotik_eoip.h"

static ng_constructor_t	ng_meoip_constructor;
static ng_rcvmsg_t	ng_meoip_rcvmsg;
static ng_shutdown_t	ng_meoip_shutdown;
static ng_newhook_t	ng_meoip_newhook;
static ng_connect_t	ng_meoip_connect;
static ng_rcvdata_t	ng_meoip_rcvdata;
static ng_disconnect_t	ng_meoip_disconnect;


static const struct ng_parse_struct_field ng_meoip_stats_type_fields[] =
	NG_MIKROTIK_EOIP_STATS_TYPE_INFO;
static const struct ng_parse_type ng_meoip_stats_type =
{
	&ng_parse_struct_type,
	&ng_meoip_stats_type_fields
};

static const struct ng_cmdlist ng_meoip_cmdlist[] =
{
	{
		NGM_MIKROTIK_EOIP_COOKIE,
		NGM_MIKROTIK_EOIP_GET_GLOBAL_STATS,
		"getglobalstats",
		NULL,
		&ng_meoip_stats_type
	},
	{
		NGM_MIKROTIK_EOIP_COOKIE,
		NGM_MIKROTIK_EOIP_GET_TUNNEL_STATS,
		"gettunnelstats",
		&ng_parse_uint16_type,
		&ng_meoip_stats_type
	},
	{ 0 }
};

static struct ng_type typestruct =
{
	.version =	NG_ABI_VERSION,
	.name =	NG_MIKROTIK_EOIP_NODE_TYPE,
	.constructor =	ng_meoip_constructor,
	.rcvmsg =	ng_meoip_rcvmsg,
	.shutdown =	ng_meoip_shutdown,
	.newhook =	ng_meoip_newhook,
	.connect =	ng_meoip_connect,
	.rcvdata = 	ng_meoip_rcvdata,
	.disconnect =	ng_meoip_disconnect,
	.cmdlist = 	ng_meoip_cmdlist
};

NETGRAPH_INIT(mikrotik_eoip, &typestruct);

struct tunnel_info
{
	uint16_t	id;
	uint32_t	packets_in;
	uint32_t	packets_out;
	uint64_t	bytes_in;
	uint64_t	bytes_out;
	hook_p	hook;
};

struct node_priv
{
	node_p	self;
	hook_p	out;
	uint32_t	packets_in;
	uint32_t	packets_out;
	uint64_t	bytes_in;
	uint64_t	bytes_out;
#ifdef	NO_LINEAR_HOOK_LOOKUP
	hook_p	hooks[0x10000];
#endif
};

typedef struct node_priv *priv_p;

/* stolen from ng_pptpgre.c */
/* This is not really GRE header, since it splits 'key' field into 2 fields:
 * length (in network byte order) and tunnel ID (in little-endian byte order).
 * Also, other optional fields are not used */
struct greheader {
#if BYTE_ORDER == LITTLE_ENDIAN
	u_char	recursion:3;	/* recursion control */
	u_char	ssr:1;	/* strict source route */
	u_char	hasSeq:1;	/* sequence number present */
	u_char	hasKey:1;	/* key present */
	u_char	hasRoute:1;	/* routing present */
	u_char	hasSum:1;	/* checksum present */
	u_char	vers:3;	/* version */
	u_char	flags:4;	/* flags */
	u_char	hasAck:1;	/* acknowlege number present */
#elif BYTE_ORDER == BIG_ENDIAN
	u_char	hasSum:1;	/* checksum present */
	u_char	hasRoute:1;	/* routing present */
	u_char	hasKey:1;	/* key present */
	u_char	hasSeq:1;	/* sequence number present */
	u_char	ssr:1;	/* strict source route */
	u_char	recursion:3;	/* recursion control */
	u_char	hasAck:1;	/* acknowlege number present */
	u_char	flags:4;	/* flags */
	u_char	vers:3;	/* version */
#else
#error BYTE_ORDER is not defined properly
#endif
	u_int16_t proto; /* protocol (ethertype) */
	u_int16_t length; /* payload length */
	u_int16_t tunnelid; /* Mikrotik's tunnel id*/
};

static struct greheader prefilled_header =
{
	.hasKey = 1,
	.vers = 1,
#if BYTE_ORDER == LITTLE_ENDIAN
	.proto = 0x0064
#elif BYTE_ORDER == BIG_ENDIAN
	.proto = 0x6400
#endif
};

static int
ng_meoip_constructor(node_p node)
{
	priv_p	privdata;

	privdata = malloc(sizeof(*privdata), M_NETGRAPH, M_WAITOK | M_ZERO);

	NG_NODE_SET_PRIVATE(node, privdata);
	privdata->self = node;

	return 0;
}

static int
ng_meoip_shutdown(node_p node)
{
	const priv_p privdata = NG_NODE_PRIVATE(node);

	NG_NODE_SET_PRIVATE(node, NULL);
	NG_NODE_UNREF(node);
	if (privdata != NULL)
		free(privdata, M_NETGRAPH);

	return 0;
}

#ifndef NO_LINEAR_HOOK_LOOKUP
static ng_fn_eachhook	check_id;

static int
check_id(hook_p hook, void *id)
{
	struct tunnel_info *info = NG_HOOK_PRIVATE(hook);

	if (info != NULL && info->id == (uint16_t)(*((unsigned int *)id)))
		return 0;
	else
		return 1;
}

static hook_p
find_hook(node_p node, uint16_t id)
{
	hook_p result = NULL;
	unsigned int val = id;

	NG_NODE_FOREACH_HOOK(node, check_id, (void*)&val, result);

	return result;
}
#endif /* NO_LINEAR_HOOK_LOOKUP */

static int
ng_meoip_newhook(node_p node, hook_p hook, const char *name)
{
	const priv_p priv = NG_NODE_PRIVATE(node);
	const char *p;
	unsigned int id = 0;
	struct tunnel_info *info;

	if (strncmp(name, NGM_MIKROTIK_EOIP_HOOK_OUT,
				strlen(NGM_MIKROTIK_EOIP_HOOK_OUT)) == 0)
	{
		priv->out = hook;
		return (0);
	}

	/* check that hook name contains only digits */
	p = name;
	while(*p != '\0')
	{
		if (!isdigit(*p))
			return (EINVAL);
		id *= 10;
		id += (*p) - '0';
		if (id > 0xffff)
			return (EINVAL);
		p++;
	}

	/* check if id is used */
#ifndef NO_LINEAR_HOOK_LOOKUP
	if (find_hook(node, id) != NULL)
		return (EADDRINUSE);
#else
	if (priv->hooks[id] != NULL)
		return (EADDRINUSE);

	priv->hooks[id] = hook;
#endif /* NO_LINEAR_HOOK_LOOKUP */

	info = malloc(sizeof(*info), M_NETGRAPH, M_WAITOK | M_ZERO);
	NG_HOOK_SET_PRIVATE(hook, info);
	info->id = id;
	info->hook = hook;
	return (0);
}

static int
ng_meoip_connect(hook_p hook)
{
	return (0);
}

static int
ng_meoip_disconnect(hook_p hook)
{
	struct tunnel_info *info = NG_HOOK_PRIVATE(hook);
	node_p node = NG_HOOK_NODE(hook);

	if (info != NULL)
	{
#ifdef NO_LINEAR_HOOK_LOOKUP
		((priv_p)NG_NODE_PRIVATE(node))->hooks[info->id] = NULL;
#endif
		free(info, M_NETGRAPH);
	} else
		((priv_p)NG_NODE_PRIVATE(node))->out = NULL;

	if ((NG_NODE_NUMHOOKS(node)) == 0
			&& NG_NODE_IS_VALID(node))
		ng_rmnode_self(node);
	return (0);
}

static int
ng_meoip_rcvmsg(node_p node, item_p item, hook_p hook)
{
	int err = 0;
	struct ng_mesg *msg, *resp = NULL;
	struct ng_mikrotik_eoip_stats *stats;
	struct tunnel_info *info;
	priv_p priv = NG_NODE_PRIVATE(node);
	hook_p thook;
	unsigned int id;

	NGI_GET_MSG(item, msg);

	if (msg->header.typecookie != NGM_MIKROTIK_EOIP_COOKIE)
	{
		err = EINVAL;
		goto done;
	}

	switch (msg->header.cmd)
	{
	case NGM_MIKROTIK_EOIP_GET_GLOBAL_STATS:
		NG_MKRESPONSE(resp, msg, sizeof(*stats), M_WAITOK);
		stats = (struct ng_mikrotik_eoip_stats *)resp->data;
		stats->packets_in = priv->packets_in;
		stats->packets_out = priv->packets_out;
		stats->bytes_in = priv->bytes_in;
		stats->bytes_out = priv->bytes_out;
		break;
	case NGM_MIKROTIK_EOIP_GET_TUNNEL_STATS:
		if (msg->header.arglen != sizeof(u_int16_t))
		{
			err = EINVAL;
			break;
		}

		id = *((uint16_t*)msg->data);

#ifndef NO_LINEAR_HOOK_LOOKUP
		if ((thook = find_hook(node, id)) == NULL)
		{
			err = EINVAL;
			break;
		}
#else
		thook = priv->hooks[id];
#endif

		info = NG_HOOK_PRIVATE(thook);

		NG_MKRESPONSE(resp, msg, sizeof(*stats), M_WAITOK);
		stats = (struct ng_mikrotik_eoip_stats *)resp->data;
		stats->packets_in = info->packets_in;
		stats->packets_out = info->packets_out;
		stats->bytes_in = info->bytes_in;
		stats->bytes_out = info->bytes_out;
		break;
	default:
		err = EINVAL;
		break;
	}

done:
	NG_RESPOND_MSG(err, node, item, resp);
	NG_FREE_MSG(msg);

	return err;
}

static int
ng_meoip_rcvdata(hook_p hook, item_p item)
{
	int err = 0;
	struct mbuf *m;
	node_p node = NG_HOOK_NODE(hook);
	priv_p priv = NG_NODE_PRIVATE(node);
	struct tunnel_info *info = NG_HOOK_PRIVATE(hook);
	struct greheader *gre;
	size_t size;
	unsigned int id;
	struct ip *ip;
	hook_p outhook;

	NGI_GET_M(item, m);
	if (info == NULL) /* received on "out" */
	{
		if (m->m_pkthdr.len < sizeof(*ip) + sizeof(*gre))
		{
			err = EINVAL;
			goto done;
		}

		if ((m = m_pullup(m, sizeof(*ip) + sizeof(*gre))) == NULL)
		{
			err = ENOBUFS;
			goto done;
		}

		ip = mtod(m, struct ip *);
		size = ip->ip_hl << 2;
		if (size != sizeof(*ip))
		{
			/* IP header has additional options */
			if ((m = m_pullup(m, size + sizeof(*gre))) == NULL)
			{
				err = ENOBUFS;
				goto done;
			}
		}

		priv->packets_in++;
		priv->bytes_in += m->m_pkthdr.len - size - sizeof(*gre);

		/* find out to which hook we should forward frame */
		gre = (struct greheader *)((uint8_t*)ip + size);
		id = le16toh(gre->tunnelid);

#ifndef NO_LINEAR_HOOK_LOOKUP
		if ((outhook = (find_hook(node, id))) == NULL)
#else
		if ((outhook = priv->hooks[id]) == NULL)
#endif
		{
			err = ENETUNREACH;
			goto done;
		}

		/* some additional checks */
		if (gre->proto != prefilled_header.proto)
		{
			err = EINVAL;
			goto done;
		}

		/* Cut IP and GRE headers */
		m_adj(m, size + sizeof(*gre));
		size = m->m_pkthdr.len;
		NG_FWD_NEW_DATA(err, item, outhook, m);
		info = NG_HOOK_PRIVATE(outhook);
		info->packets_in++;
		info->bytes_in += size;
		return (err);
	}
	else
	{
		/* Forward to ng_ksocket hook */
		if (priv->out == NULL)
		{
			err = ENETUNREACH;
			goto done;
		}

		size = m->m_pkthdr.len;
		M_PREPEND(m, sizeof(*gre), M_NOWAIT);
		if ((m == NULL) || ((m = m_pullup(m, sizeof(*gre))) == NULL))
		{
			err = ENOBUFS;
			goto done;
		}

		/* Prepare eoip header */
		gre = mtod(m, struct greheader *);
		bcopy(&prefilled_header, gre, sizeof(*gre));
		gre->length = htons(size);
		gre->tunnelid = htole16(info->id);

		NG_FWD_NEW_DATA(err, item, priv->out, m);
		priv->packets_out++;
		info->packets_out++;
		priv->bytes_out += size;
		info->bytes_out += size;
		return (err);
	}

done:
	NG_FREE_ITEM(item);
	if (m != NULL)
		NG_FREE_M(m);
	return (err);
}
